

from .basemixin import BaseMixin
from .log import NodeLoggingMixin


class NodeBusyMixin(NodeLoggingMixin, BaseMixin):
    def __init__(self, *args, **kwargs):
        super(NodeBusyMixin, self).__init__(*args, **kwargs)
        self._busy = False

    @property
    def busy(self):
        if self._busy > 0:
            return True
        else:
            return False

    def busy_set(self):
        self._busy += 1

    def busy_clear(self):
        self._busy -= 1
        if self._busy < 0:
            self.log.warn("Busy cleared too many times!")
            self._busy = 0

    def _busy_setter(self, value):
        self.log.debug("Setting node busy status to {0}".format(value))
        self._busy = value





import os
import appdirs

from .nodeid import NodeIDMixin
from .log import NodeLoggingMixin
from .busy import NodeBusyMixin
from .shell import BaseShellMixin
from .http import HttpClientMixin
from .resources import ResourceManagerMixin
from .background import BackgroundCoreMixin


class BaseIoTNode(BackgroundCoreMixin,
                  ResourceManagerMixin,
                  HttpClientMixin,
                  BaseShellMixin,
                  NodeBusyMixin,
                  NodeLoggingMixin,
                  NodeIDMixin):
    _has_gui = False

    def __init__(self, *args, **kwargs):
        super(BaseIoTNode, self).__init__(*args, **kwargs)

    def install(self):
        super(BaseIoTNode, self).install()
        self.log.info("Installing Node with ID {id}", id=self.id)
        os.makedirs(appdirs.user_config_dir(self.config.appname), exist_ok=True)

    def start(self):
        super(BaseIoTNode, self).start()
        self.log.info("Starting Node with ID {id}", id=self.id)

    def stop(self):
        super(BaseIoTNode, self).stop()
        self.log.info("Stopping Node with ID {id}", id=self.id)

    def exit(self):
        self.stop()





import os
import io
import sys
import time
from twisted import logger
from twisted.logger import LogLevel
from twisted.logger import LogLevelFilterPredicate
from twisted.logger import FilteringLogObserver
from twisted.logger import textFileLogObserver

from appdirs import user_log_dir
from datetime import datetime

from .config import ElementSpec, ItemSpec
from .config import ConfigMixin
from .basemixin import BaseMixin

import logging
logging.basicConfig(level=logging.INFO)


class NodeLoggingMixin(ConfigMixin, BaseMixin):
    _log = None

    def __init__(self, *args, **kwargs):
        super(NodeLoggingMixin, self).__init__(*args, **kwargs)
        self._log_file = None
        self.log_prune()
        self._log = logger.Logger(namespace=self.appname,
                                  source=self)
        self.reactor.callWhenRunning(self._start_logging)

    def install(self):
        super(NodeLoggingMixin, self).install()
        _elements = {
            'debug': ElementSpec('debug', 'debug', ItemSpec(bool, fallback=False)),
        }

        for element, element_spec in _elements.items():
            self.config.register_element(element, element_spec)

    def _observers(self):
        if self.config.debug:
            level = LogLevel.debug
        else:
            level = LogLevel.info

        return [
            # STDLibLogObserver(),
            FilteringLogObserver(
                textFileLogObserver(sys.stdout),
                predicates=[LogLevelFilterPredicate(LogLevel.warn)]
            ),
            FilteringLogObserver(
                textFileLogObserver(io.open(self.log_file, 'a')),
                predicates=[LogLevelFilterPredicate(level)]
            ),
        ]

    def _start_logging(self):
        # TODO Mention that docs don't say reactor should be running
        # TODO Mention that docs are confusing about how extract works
        # TODO Find out about a functional print to console observer
        # TODO Mention problem with IOBase vs TextIOWrapper
        # TODO log_source is not set when logger instantiated in __init__
        logger.globalLogBeginner.beginLoggingTo(self._observers())
        self.log.info("Logging to {logfile}", logfile=self.log_file)

    @property
    def log(self):
        return self._log

    @property
    def log_file(self):
        if not self._log_file:
            self._log_file = os.path.join(
                self.log_dir,
                'runlog_{0}'.format(datetime.today().strftime('%d%m%y'))
            )
        return self._log_file

    def log_prune(self):
        for fname in self.log_files:
            fpath = os.path.join(self.log_dir, fname)
            mtime = os.path.getmtime(fpath)
            if time.time() - mtime > (7 * 24 * 60 * 60):
                os.remove(fpath)

    @property
    def log_files(self):
        return self._log_files()

    def _log_files(self):
        for filename in os.listdir(self.log_dir):
            if os.path.isfile(os.path.join(self.log_dir, filename)):
                yield filename

    @property
    def log_dir(self):
        os.makedirs(user_log_dir(self.appname), exist_ok=True)
        return user_log_dir(self.appname)





import os
from twisted import logger

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound


class GenericSequencePersistenceManager(object):
    _db_name = 'generic'
    _db_model = None
    _db_metadata = None

    def __init__(self, actual):
        self._actual = actual
        self._items = []
        self._is_loaded = False
        self._log = None

        self._db_engine = None
        self._db = None
        self._db_dir = None
        _ = self.db

    @property
    def log(self):
        if not self._log:
            self._log = logger.Logger(
                namespace="{0}.pm".format(self.__class__.__name__),
                source=self
            )
        return self._log

    def get(self, reload=True):
        self._persistence_load(force=reload)
        return self._items

    def update(self, items):
        self._is_loaded = False
        self.clear()
        if not len(items):
            return
        session = self.db()
        self.log.debug("Inserting Items to DB '{}'"
                       "".format(self._db_name))
        try:
            for idx, item in enumerate(items):
                robj = self._db_model(idx, item)
                self._insert_item(idx, item)
                session.add(robj)
            session.commit()
            session.flush()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def _insert_item(self, seq, item):
        pass

    def _clear_item(self, item):
        pass

    def clear(self):
        self._is_loaded = False
        session = self.db()
        self.log.debug("Clearing Persistent Items from DB '{}'"
                       "".format(self._db_name))
        try:
            results = self.db_get_resources(session).all()
        except NoResultFound:
            session.close()
            return
        try:
            for robj in results:
                session.delete(robj)
                self._clear_item(robj)
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def _persistence_load(self, force=False):
        if self._is_loaded and not force:
            return
        self.log.debug("Loading Persistent Items from DB '{}'"
                       "".format(self._db_name))
        session = self.db()
        try:
            results = self.db_get_resources(session).all()
        except NoResultFound:
            session.close()
            self._items = []
        try:
            _items = []
            for robj in results:
                _items.append(robj.native())
        finally:
            session.close()
        self._items = _items
        self.log.info("Got {} Items from DB '{}'."
                      "".format(len(self._items), self._db_name))

    def db_get_resources(self, session, seq=None):
        q = session.query(self.db_model)
        if seq is not None:
            q = q.filter(
                self.db_model.seq == seq
            )
        else:
            q = q.order_by(self.db_model.seq)
        return q

    @property
    def db_model(self):
        if not self._db_model:
            raise AttributeError("_db_model Needs to be specified")
        return self._db_model

    @property
    def db_metadata(self):
        if not self._db_metadata:
            raise AttributeError("_db_metadata needs to be provided")
        return self._db_metadata

    @property
    def db(self):
        if self._db is None:
            self._db_engine = create_engine(self.db_url)
            self.db_metadata.create_all(self._db_engine)
            self._db = sessionmaker(expire_on_commit=False)
            self._db.configure(bind=self._db_engine)
        return self._db

    @property
    def db_url(self):
        return 'sqlite:///{0}'.format(
            os.path.join(self.db_dir, "{}.db".format(self._db_name))
        )

    @property
    def db_dir(self):
        return self._actual.db_dir

    def render(self):
        print("----------")
        print("{} DB Content".format(self._db_name))
        print("----------")
        session = self.db()
        try:
            results = self.db_get_resources(session).all()
        finally:
            session.close()
        print(results)
        print("----------")





import uuid
import netifaces

from .basemixin import BaseMixin
from .config import ConfigMixin
from .config import ElementSpec, ItemSpec


class NodeIDMixin(ConfigMixin, BaseMixin):
    _node_id_netifaces_fallback_interfaces = ['wlp1s0', 'wlan0', 'wlo1', 'eth0']

    def __init__(self, *args, **kwargs):
        super(NodeIDMixin, self).__init__(*args, **kwargs)
        self._id = None

    def install(self):
        super(NodeIDMixin, self).install()
        _elements = {
            'node_id_getter': ElementSpec('id', 'getter', ItemSpec(fallback='netifaces')),
            'node_id_interface': ElementSpec('id', 'interface', ItemSpec(fallback=None)),
            'node_id_override': ElementSpec('id', 'override', ItemSpec(fallback=None)),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    @property
    def id(self):
        if self._id is None:
            self._id = self._get_id()
        return self._id

    def _get_id(self):
        if self.config.node_id_override is not None:
            return self.config.node_id_override
        getter = "_get_node_id_{0}".format(self.config.node_id_getter)
        params = {'interface': self.config.node_id_interface}
        return getattr(self, getter)(**params).upper()

    def _get_node_id_uuid(self, **_):
        node_id = uuid.getnode()
        if (node_id >> 40) % 2:
            raise OSError("The system does not seem to have a valid MAC")
        return hex(node_id)[2:]

    def _get_node_id_netifaces_guess(self):
        fallback_interfaces = self._node_id_netifaces_fallback_interfaces
        available_interfaces = netifaces.interfaces()
        default_gateway = netifaces.gateways()['default']
        if default_gateway:
            return default_gateway[netifaces.AF_INET][1]
        for iface in fallback_interfaces:
            if iface in available_interfaces:
                return iface

    def _get_node_id_netifaces(self, **kwargs):
        interface = kwargs.get('interface', None)
        if interface is None:
            interface = self._get_node_id_netifaces_guess()
        mac = netifaces.ifaddresses(interface)[netifaces.AF_LINK][0]['addr']
        return mac.replace(':', '')





import os
import time
import psutil
from datetime import datetime
from datetime import timedelta
from functools import partial

from twisted import logger
from twisted.internet.defer import succeed
from twisted.internet.task import cooperate
from twisted.web.client import ResponseFailed

from sqlalchemy import Column
from sqlalchemy import Integer
from sqlalchemy import Text
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.exc import NoResultFound

from .http import HttpClientMixin
from .http import _http_errors
from .config import ElementSpec, ItemSpec

from .constants import ASSET
from .constants import CONTENT

Base = declarative_base()
metadata = Base.metadata


class ResourceModel(Base):
    __tablename__ = 'resources'

    id = Column(Integer, primary_key=True)
    filename = Column(Text, index=True)
    url = Column(Text)
    rtype = Column(Integer)


class CacheableResource(object):
    def __init__(self, manager, filename, url=None, rtype=None):
        self._manager = manager
        self._filename = filename
        self._url = url
        self._rtype = rtype
        self._cache_path = None
        if not self._rtype:
            self.load()

    @property
    def filename(self):
        return self._filename

    @property
    def url(self):
        return self._url

    @property
    def rtype(self):
        return self._rtype

    @rtype.setter
    def rtype(self, value):
        if not value:
            value = 0
        if value in [0, ASSET, CONTENT]:
            self._rtype = value

    @property
    def cache_path(self):
        if not self._cache_path:
            self._cache_path = self._manager.cache_path(self.filename)
        return self._cache_path

    @property
    def filepath(self):
        return self.cache_path

    @property
    def available(self):
        return os.path.exists(self.cache_path)

    @property
    def is_asset(self):
        if self.rtype == ASSET:
            return True
        else:
            return False

    @property
    def is_content(self):
        if self.rtype == CONTENT:
            return True
        else:
            return False

    @property
    def is_orphaned(self):
        if not self.rtype:
            return True
        else:
            return False

    def commit(self):
        session = self._manager.db()
        try:
            try:
                robj = session.query(ResourceModel).filter_by(filename=self.filename).one()
            except NoResultFound:
                robj = ResourceModel()
                robj.filename = self.filename

            robj.url = self.url
            robj.rtype = self.rtype

            session.add(robj)
            session.flush()
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def load(self):
        session = self._manager.db()
        try:
            robj = session.query(ResourceModel).filter_by(filename=self.filename).one()
            self._url = robj.url
            self._rtype = robj.rtype
        except:
            session.rollback()
            raise
        finally:
            session.close()

    @property
    def node(self):
        return self._manager.node

    def __repr__(self):
        return "{1} {0}".format(self.filename, self.rtype)


class ResourceManager(object):
    def __init__(self, node, **kwargs):
        self._resource_class = kwargs.pop('resource_class', CacheableResource)
        self._log = None
        self._node = node
        self._db = None
        self._db_engine = None
        self._db_dir = None
        self._cache_dir = None
        self._active_downloads = []
        super(ResourceManager, self).__init__(**kwargs)

    @property
    def node(self):
        return self._node

    @property
    def log(self):
        if not self._log:
            self._log = logger.Logger(namespace="rm", source=self)
        return self._log

    def has(self, filename):
        # Check if a resource is in defined by the manager.
        # This makes no guarantees about it existing in the cache.
        session = self.db()
        try:
            _ = session.query(ResourceModel).filter_by(filename=filename).one()
            return True
        except NoResultFound:
            return False
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def get(self, filename):
        # Get the resource object bound to the manager.
        # This makes no guarantees about it existing in the cache.
        # Check the available property of the returned resource to know if
        # it is there.
        return self._resource_class(self, filename)

    def insert(self, filename, url=None, rtype=CONTENT):
        # Create a resource object and insert it into the manager.
        # This makes no guarantees about it existing in the cache.
        resource = self._resource_class(self, filename, url, rtype)
        resource.commit()

    def remove(self, filename):
        session = self.db()
        # print("Trying to remove {0} from rdb".format(filename))
        try:
            try:
                robj = session.query(ResourceModel).filter_by(filename=filename).one()
            except NoResultFound:
                return
            session.delete(robj)
            # print("Committing rdel")
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def prefetch(self, resource, retries=None, semaphore=None):
        # Given a resource belonging to this resource manager, download it
        # to the cache if it isn't already there or update its mtime if it is.
        if resource.filename in self._active_downloads:
            return
        if resource.available:
            with open(resource.cache_path, 'a'):
                os.utime(resource.cache_path, None)
            return

        if retries is None:
            retries = self._node.config.resource_prefetch_retries

        d = self._fetch(resource, semaphore=semaphore)

        def _retry(failure, attempts=1):
            failure.trap(ResponseFailed, *_http_errors)
            attempts = attempts - 1
            if attempts:
                self._node.reactor.callLater(
                    self._node.config.resource_prefetch_retry_delay,
                    self.prefetch, resource, retries=attempts
                )
        d.addErrback(partial(_retry, attempts=retries))
        return d

    def _fetch(self, resource, semaphore=None):
        self._active_downloads.append(resource.filename)
        self.log.debug("Requesting download of {filename}", filename=resource.filename)
        d = self._node.http_download(resource.url, resource.cache_path, semaphore=semaphore)

        # Update timestamps for the downloaded file to reflect start of
        # download instead of end. Consider if this is wise.
        def _dl_finalize(r, times, _):
            with open(r.cache_path, 'a'):
                os.utime(r.cache_path, times)

        d.addCallback(
            partial(_dl_finalize, resource, (time.time(), time.time()))
        )

        def _vacate_download(maybe_failure):
            self._active_downloads.remove(resource.filename)
            return maybe_failure
        d.addBoth(_vacate_download)
        return d

    @property
    def db(self):
        if self._db is None:
            self._db_engine = create_engine(self.db_url)
            metadata.create_all(self._db_engine)
            self._db = sessionmaker(expire_on_commit=False)
            self._db.configure(bind=self._db_engine)
        return self._db

    @property
    def db_url(self):
        return 'sqlite:///{0}'.format(os.path.join(self.db_dir, 'resources.db'))

    @property
    def db_dir(self):
        return self._node.db_dir

    @property
    def cache_dir(self):
        return self._node.cache_dir


class NothingToTrimError(Exception):
    pass


class CachingResourceManager(ResourceManager):
    _excluded_folders = ['log']

    def __init__(self, *args, **kwargs):
        super(CachingResourceManager, self).__init__(*args, **kwargs)
        self.cache_max_size = self._node.config.cache_max_size

    def prefetch(self, resource, retries=None, semaphore=None):
        # When done, trim the cache.
        d = super(CachingResourceManager, self).prefetch(
            resource, retries=retries, semaphore=semaphore
        )
        if d:
            def fetch_postprocess(_):
                task = cooperate(
                    self.cache_trim()
                )
                td = task.whenDone()

                def _report_done(_):
                    # self.log.debug("Cache trim complete.")
                    pass
                td.addCallback(_report_done)
            d.addCallback(fetch_postprocess)
        else:
            d = succeed(True)
        return d

    def cache_remove(self, filename):
        size = self.cache_file_size(filename)
        # self.log.debug("Removing {filename} of size {size} from cache",
        #                filename=filename, size=size)
        try:
            os.remove(self.cache_path(filename))
        except FileNotFoundError:
            pass
        return size

    def cache_has(self, filename):
        r = os.path.exists(self.cache_path(filename))
        # if r:
        #     self.log.debug("{filename} found in the cache",
        #                    filename=filename)
        return r

    def cache_path(self, filename):
        return os.path.join(self.cache_dir, filename)

    def cache_file_size(self, filename):
        try:
            rv = os.path.getsize(self.cache_path(filename))
            return rv
        except OSError:
            return 0

    @property
    def cache_files(self):
        return self._cache_files()

    def _cache_files(self):
        for filename in os.listdir(self.cache_dir):
            if os.path.isfile(self.cache_path(filename)) and \
                    not filename.endswith('.partial'):
                yield filename

    @property
    def cache_size(self):
        return sum(map(self.cache_file_size, self.cache_files))

    def cache_clear(self):
        raise NotImplementedError

    def cache_trim(self, max_size=None, space_for=0):
        # Trim the cache cache down to max_size by removing content items to
        # the provided max_size.
        #  - First remove all cache items which are orphaned (aren't in the
        #    resource database) one by one. Note that items are added to the
        #    database before any attempt is made to prefetch it.
        #  - Remove cache items which are defined as content by its rtype as
        #    per the auto-selected trimfunc.
        #
        # fifo
        #  - Selected if both 'next_use' and 'last_use' are not defined on
        #    the resource.
        #  - Remove the oldest cached content file by mtime, one by one.
        #  - Note that this implementation actually modifies a typical FIFO
        #    cache into a pseudo-LRU cache by it's updating cache item
        #    timestamps whenever prefetch is called.
        #
        # lru
        #  - Selected if the resource defines 'last_use' and not 'next_use'.
        #  - Remove the least recently used cached content file, one by one.
        #  - Note that if this is used, the application should be sure to set
        #    'last_use' to something meaningful, perhaps timestamp of
        #    creation time, before trim is called.
        #  - LRU is not intended for regular use, it's here for largely
        #    academic purposes.
        #
        # predictive
        #  - Selected if the resource defines 'next_use'. This is the
        #    preferred cache trimmer.
        #  - Remove cached content items which have no known 'next_use',
        #    one by one
        #  - Remove cached content with 'next_use' set to the past.
        #  - Remove cached content items with 'next_use' most in the future,
        #    up to about 30 minutes from the current time
        #
        if max_size is None:
            max_size = self.cache_max_size
        max_size = max_size - space_for
        if hasattr(self._resource_class, 'next_use'):
            trimmer = self._cache_trimmer_predictive
        elif hasattr(self._resource_class, 'last_use'):
            trimmer = self._cache_trimmer_lru
        else:
            trimmer = self._cache_trimmer_fifo
        current_size = self.cache_size
        # self.log.debug("Attempting to trim cache to {max_size} from "
        #                "{current_size} with {trimmer}",
        #                max_size=max_size, current_size=current_size,
        #                trimmer=trimmer.__name__)
        if current_size > max_size:
            r = list(self.cache_resources)
            resources = [(x, x.next_use) for x in r]
            while current_size > max_size:
                try:
                    resources, rv = self._cache_trimmer(resources, trimmer)
                    current_size -= rv
                except NothingToTrimError:
                    break
                yield None

    @property
    def cache_resources(self):
        return self._cache_resources()

    def _cache_resources(self):
        for filename in self.cache_files:
            if self.node.cache_trim_exclusions and \
                    filename in self.node.cache_trim_exclusions:
                continue
            resource = self.get(filename)
            yield resource

    def _cache_cremove(self, resources, r):
        resources.remove(r)
        return resources, self.cache_remove(r[0].filename)

    def _cache_trimmer(self, resources, trimfunc):
        # TODO Check about dangling temporary files.
        for r in resources:
            if r[0].is_orphaned:
                return self._cache_cremove(resources, r)
        return trimfunc([x for x in resources if x[0].is_content])

    def _cache_trimmer_fifo(self, resources):
        r = sorted(resources, key=lambda x: os.path.getmtime(x[0].cache_path))
        # self._cache_debug(r, 'by mtime',
        #                   lambda x: os.path.getmtime(x.cache_path))
        if len(r):
            return self._cache_cremove(resources, r[0])
        else:
            raise NothingToTrimError()

    def _cache_trimmer_lru(self, resources):
        r = sorted(resources, key=lambda x: x[1])
        # self._cache_debug(r, 'by last_use', lambda x: x.last_use)
        if len(r):
            return self._cache_cremove(resources, r[0])
        else:
            raise NothingToTrimError()

    def _cache_trimmer_predictive(self, resources):
        # No next_use
        r = [x for x in resources if not x[1]]
        # self._cache_debug(r, 'by no next_use', lambda x: x.next_use)
        if len(r):
            # self.log.debug("NO NEXT USE")
            return self._cache_trimmer_fifo(r)

        # Next_use, next_use in the past
        cutoff = datetime.now()
        r = sorted((x for x in resources if x[1] < cutoff),
                   key=lambda x: x[1])
        # self._cache_debug(r, 'by next_use in past', lambda x: x.next_use)
        if len(r):
            # self.log.debug('NEXT USE IN PAST')
            return self._cache_cremove(resources, r[0])

        # Next_use, next_use in the future
        cutoff += timedelta(minutes=20)
        r = sorted((x for x in resources if x[1] > cutoff),
                   key=lambda x: x[1], reverse=True)
        # self._cache_debug(r, 'by next_use in future', lambda x: x.next_use)
        if len(r):
            # self.log.debug('NEXT USE IN FUTURE')
            return self._cache_cremove(resources, r[0])

        raise NothingToTrimError()

    def _cache_debug(self, resources, title, keyfunc):
        self.log.debug("------------------------------------")
        self.log.debug("Cache Content {0}".format(title))
        for r, _ in resources:
            self._node.log.debug(
                "{key} {filename}",
                filename=r.filename,
                key=keyfunc(r)
            )
        self.log.debug("----------------------------------- ")


class ResourceManagerMixin(HttpClientMixin):
    def __init__(self, *args, **kwargs):
        self._resource_manager = None
        self._resource_class = kwargs.pop('resource_class', CacheableResource)
        super(ResourceManagerMixin, self).__init__(*args, **kwargs)

    def install(self):
        super(ResourceManagerMixin, self).install()
        _default_cache_size = int((psutil.disk_usage('/').total - 5000000000) * 0.7)
        _elements = {
            'resource_prefetch_retries': ElementSpec('resources', 'prefetch_retries', ItemSpec(int, fallback=6)),
            'resource_prefetch_retry_delay': ElementSpec('resources', 'prefetch_retry_delay', ItemSpec(int, fallback=60)),
            'cache_max_size': ElementSpec('cache', 'max_size', ItemSpec(int, fallback=_default_cache_size))
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    @property
    def resource_manager(self):
        if not self._resource_manager:
            self.log.info("Initializing resource manager")
            self._resource_manager = CachingResourceManager(
                self, resource_class=self._resource_class,
            )
        return self._resource_manager

    @property
    def cache_trim_exclusions(self):
        return []





import os
import time
from collections import namedtuple
from six.moves.urllib.parse import urlparse

from ebs.linuxnode.core.basemixin import BaseMixin
from ebs.linuxnode.core.log import NodeLoggingMixin
from ebs.linuxnode.core.resources import ResourceManagerMixin
from ebs.linuxnode.core.config import ElementSpec, ItemSpec


# TODO This is here, but not particularly tested. Expect it to only work for video.
# In general, background sequencing should be handled outside the Background code or
# by dedicated infrastructure with effectively allows gui_bg to accept a list of this class.
BackgroundSpec = namedtuple('BackgroundSpec', ["target", "bgcolor", "callback", "duration"],
                            defaults=[None, None, None])


class BackgroundProviderBase(object):
    is_visual = True

    def __init__(self, actual):
        self._actual = actual
        self._widget = None
        self._end_call = None
        self._paused = False
        self._eresidual = None
        self._callback = None

    @property
    def actual(self):
        if hasattr(self._actual, 'actual'):
            return self._actual.actual
        else:
            return self._actual

    def check_support(self, target):
        # Check if the provider supports the target and
        # if the target exists.
        raise NotImplementedError

    def play(self, target, duration=None, callback=None, **kwargs):
        # Create a Widgetized Background and return it.
        # It will be attached later.
        if duration and callback:
            self._callback = callback
            self._end_call = self.actual.reactor.callLater(duration, callback)

    def stop(self):
        # Stop and unload the Widgetized Background.
        # The widget has already been detached.
        if self._end_call and self._end_call.active():
            self._end_call.cancel()
        self._widget = None

    def pause(self):
        # Pause the Widgetized Background.
        # It has already been detached.
        if self._paused:
            return
        if self._end_call and self._end_call.active():
            ietime = self._end_call.getTime()
            ptime = time.time()
            self._eresidual = ietime - ptime
            self._end_call.cancel()
        self._paused = True

    def resume(self):
        # Resume the Widgetized Background.
        # It will be attached later.
        if not self._paused:
            return
        self._paused = False
        if self._eresidual:
            self._end_call = self.actual.reactor.callLater(self._eresidual, self._callback)
            self._eresidual = None


class BackgroundCoreMixin(ResourceManagerMixin, NodeLoggingMixin, BaseMixin):
    def __init__(self, *args, **kwargs):
        super(BackgroundCoreMixin, self).__init__(*args, **kwargs)
        self._bg_providers = []
        self._bg = None
        self._bg_current = None
        self._bg_current_provider = None

    def _background_fallback(self):
        return None

    def install(self):
        super(BackgroundCoreMixin, self).install()
        _elements = {
            'background': ElementSpec('display', 'background',
                                      ItemSpec(str, read_only=False, fallback=self._background_fallback())),
        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    def install_background_provider(self, provider):
        self.log.info("Installing BG Provider {}".format(provider))
        self._bg_providers.insert(0, provider)

    def _get_provider(self, target):
        provider: BackgroundProviderBase
        provider = None
        for lprovider in self._bg_providers:
            if lprovider.check_support(target):
                provider = lprovider
                break
        return provider

    def background_set(self, target):
        if not target:
            target = None

        provider = self._get_provider(target)
        if not provider:
            self.log.warn("Provider not found for background {}. Not Setting.".format(target))
            target = None

        if target and self.config.background != target:
            old_bg = urlparse(self.config.background)
            if not old_bg.scheme:
                old_fname = os.path.basename(old_bg.path)
                if self.resource_manager.has(old_fname):
                    self.resource_manager.remove(old_fname)
            self.config.background = target

        self.bg_update()

    def bg_clear(self):
        self._bg = None
        if self._bg_current_provider:
            self._bg_current_provider.stop()

    @property
    def bg(self):
        return self._bg_current

    @bg.setter
    def bg(self, value):
        self.log.info("Setting background to {value}", value=value)
        bgcolor, callback, duration = None, None, None
        if isinstance(value, BackgroundSpec):
            value, bgcolor, callback, duration = value

        if self._bg_current == value and not callback and not duration:
            return False

        if not bgcolor:
            try:
                bgcolor = self.config.image_bgcolor
            except:
                bgcolor = 'auto'

        provider = self._get_provider(value)

        if not provider:
            self.log.warn("Provider not found for background {}".format(value))
            value = self.config.background
            provider = self._get_provider(value)
            self.log.warn("Trying to use {} instead.".format(value))

        if not provider:
            self.log.warn("Unable to display config background. Clearing from config.")
            self.config.remove('background')
            value = self.config.background
            provider = self._get_provider(value)

        self.log.debug("Using {} to show background {}".format(provider, value))
        self.bg_clear()

        self._bg_current = value
        self._bg_current_provider = provider
        self._bg = self._bg_current_provider.play(
            value, bgcolor=bgcolor, callback=callback, duration=duration
        )

        return True

    def bg_pause(self):
        self.log.debug("Pausing Background")
        if self._bg_current_provider:
            self._bg_current_provider.pause()

    def bg_resume(self):
        self.log.debug("Resuming Background")
        if self._bg_current_provider:
            self._bg_current_provider.resume()

    def bg_update(self):
        self.bg = self.config.background

    def start(self):
        super(BackgroundCoreMixin, self).start()
        # TODO This will break unless the GUI components are in place. If it needs
        #  to be used separately, an alternate approach needs to be implemented.
        # self.reactor.callLater(3, self.bg_update)

    def stop(self):
        if self._bg_current_provider:
            self._bg_current_provider.stop()
        super(BackgroundCoreMixin, self).stop()





import os
import atexit
import shutil
import tempfile
from appdirs import user_cache_dir
from appdirs import user_config_dir
from twisted.internet import reactor


class BaseMixin(object):
    def __init__(self, *args, **kwargs):
        self._reactor = kwargs.pop('reactor', reactor)
        self._cache_dir = None
        self._config_dir = None
        self._db_dir = None
        self._temp_dir = None
        super(BaseMixin, self).__init__(*args, **kwargs)

    def install(self):
        pass

    def start(self):
        pass

    def stop(self):
        pass

    def _deferred_error_passthrough(self, failure):
        return failure

    def _deferred_error_swallow(self, failure):
        return

    @property
    def reactor(self):
        return self._reactor

    @property
    def appname(self):
        return self.config.appname

    @property
    def config_dir(self):
        if not self._config_dir:
            self._config_dir = user_config_dir(self.appname)
            os.makedirs(self._config_dir, exist_ok=True)
        return self._config_dir

    @property
    def cache_dir(self):
        if not self._cache_dir:
            self._cache_dir = user_cache_dir(self.appname)
            os.makedirs(self._cache_dir, exist_ok=True)
        return self._cache_dir

    @property
    def db_dir(self):
        if not self._db_dir:
            self._db_dir = os.path.join(self.cache_dir, 'db')
            os.makedirs(self._db_dir, exist_ok=True)
        return self._db_dir

    @property
    def temp_dir(self):
        if not self._temp_dir:
            self._temp_dir = tempfile.mkdtemp()
            atexit.register(shutil.rmtree, self._temp_dir)
        return self._temp_dir

    def sysinfo_install(self):
        self.sysinfo.app.versions.register_namespace('ebs')
        self.sysinfo.app.versions.register_namespace('tendril')





from os import environ
from twisted.internet.utils import getProcessOutput
from .basemixin import BaseMixin
from .config import ConfigMixin


class BaseShellMixin(ConfigMixin, BaseMixin):
    # def __init__(self, *args, **kwargs):
    #     self._shell_processes = {}
    #     super(BaseShellMixin, self).__init__(self, *args, **kwargs)

    def _shell_execute(self, command, response_handler):
        if len(command) > 1:
            args = command[1:]
        else:
            args = []
        d = getProcessOutput(command[0], args, env=environ)
        d.addCallback(response_handler)
        return d





import os
import base64

from functools import partial
from six.moves.urllib.parse import urlparse
from twisted.web.client import Agent
from twisted.web.client import ProxyAgent
from twisted.internet.endpoints import TCP4ClientEndpoint
from twisted.internet.defer import DeferredSemaphore
from treq.client import HTTPClient

from .basemixin import BaseMixin
from .log import NodeLoggingMixin
from .busy import NodeBusyMixin

from twisted.internet.protocol import Protocol
from twisted.web.client import ResponseDone
from twisted.web.http import PotentialDataLoss
from twisted.internet.defer import Deferred, succeed

from twisted.internet.error import TimeoutError
from twisted.internet.error import DNSLookupError
from twisted.internet.error import ConnectError
from twisted.internet.error import NoRouteError
from twisted.web.client import ResponseNeverReceived
from twisted.web.client import ResponseFailed
from twisted.web.error import SchemeNotSupported

from .config import ElementSpec, ItemSpec


class HTTPError(Exception):
    def __init__(self, response):
        self.response = response


_http_errors = (HTTPError, DNSLookupError, NoRouteError,
                TimeoutError, ConnectError, ResponseNeverReceived)


def swallow_http_error(failure):
    failure.trap(*_http_errors)
    print("Swallowing HTTP Error")


class NoResumeResponseError(Exception):
    def __init__(self, code):
        self.code = code


class DefaultHeadersHttpClient(HTTPClient):
    def __init__(self, *args, **kwargs):
        self._default_headers = kwargs.pop('headers', {})
        super(DefaultHeadersHttpClient, self).__init__(*args, **kwargs)

    def get(self, url, **kwargs):
        simple_headers = kwargs.pop('headers', {})
        simple_headers.update(self._default_headers)
        kwargs['headers'] = simple_headers
        return super(DefaultHeadersHttpClient, self).get(url, **kwargs)

    def post(self, url, **kwargs):
        simple_headers = kwargs.pop('headers', {})
        simple_headers.update(self._default_headers)
        kwargs['headers'] = simple_headers
        return super(DefaultHeadersHttpClient, self).post(url, **kwargs)


class WatchfulBodyCollector(Protocol):
    def __init__(self, finished, collector, chunktimeout, reactor):
        # TODO Reimplement with twisted.protocols.policies.TimeoutMixin
        self.latest_tick = 0
        self.chunktimeout = chunktimeout
        self.reactor = reactor
        self.finished = finished
        self.collector = collector

    def dataReceived(self, data):
        self.collector(data)
        if self.chunktimeout is not None:
            self.latest_tick += 1
            self.reactor.callLater(self.chunktimeout, self.checkTimeout,
                                   self.latest_tick)

    def checkTimeout(self, tick):
        if tick == self.latest_tick:
            # We've really timed out.
            self.transport.loseConnection()

    def connectionLost(self, reason):
        self.latest_tick = 0
        if reason.check(ResponseDone):
            self.finished.callback(None)
        elif reason.check(PotentialDataLoss):
            # http://twistedmatrix.com/trac/ticket/4840
            self.finished.callback(None)
        else:
            self.finished.errback(reason)


def watchful_collect(response, collector, chunktimeout=None, reactor=None):
    if response.length == 0:
        return succeed(None)

    d = Deferred()
    response.deliverBody(
        WatchfulBodyCollector(d, collector, chunktimeout, reactor)
    )
    return d


class HttpClientMixin(NodeBusyMixin, NodeLoggingMixin, BaseMixin):
    def __init__(self, *args, **kwargs):
        self._http_headers = {}
        self._http_client = None
        self._http_semaphore = None
        self._http_semaphore_background = None
        self._http_semaphore_download = None
        super(HttpClientMixin, self).__init__(*args, **kwargs)

    def install(self):
        super(HttpClientMixin, self).install()
        _elements = {
            'http_max_concurrent_requests': ElementSpec('http', 'max_concurrent_requests', ItemSpec(int, fallback=1)),
            'http_max_background_downloads': ElementSpec('http', 'max_background_downloads', ItemSpec(int, fallback=1)),
            'http_max_concurrent_downloads': ElementSpec('http', 'max_concurrent_downloads', ItemSpec(int, fallback=1)),
            'http_proxy_host': ElementSpec('http', 'proxy_host', ItemSpec(fallback=None)),
            'http_proxy_port': ElementSpec('http', 'proxy_port', ItemSpec(int, fallback=0)),
            'http_proxy_user': ElementSpec('http', 'proxy_user', ItemSpec(fallback=None)),
            'http_proxy_pass': ElementSpec('http', 'proxy_pass', ItemSpec(fallback=None)),
            'http_proxy_enabled': ElementSpec('_derived', self._http_proxy_enabled),
            'http_proxy_auth': ElementSpec('_derived', self._http_proxy_auth),
            'http_proxy_url': ElementSpec('_derived', self._http_proxy_url),

        }
        for name, spec in _elements.items():
            self.config.register_element(name, spec)

    def _http_proxy_enabled(self, config):
        return config.http_proxy_host is not None

    def _http_proxy_auth(self, config):
        if not config.http_proxy_user:
            return None
        if not config.http_proxy_pass:
            return config.http_proxy_user
        return "{0}:{1}".format(config.http_proxy_user, config.http_proxy_pass)

    def _http_proxy_url(self, config):
        url = config.http_proxy_host
        if config.http_proxy_port:
            url = "{0}:{1}".format(url, config.http_proxy_port)
        if config.http_proxy_auth:
            url = "{0}@{1}".format(config.http_proxy_auth, url)
        return url

    def http_get(self, url, **kwargs):
        self.log.debug("Executing HTTP GET Request\n"
                       " to URL {url}\n"
                       " with kwargs {kwargs}",
                       url=url, kwargs=kwargs)
        deferred_response = self.http_semaphore.run(
            self.http_client.get, url, **kwargs
        )
        deferred_response.addCallbacks(
            self._http_check_response,
            partial(self._http_error_handler, url=url)
        )
        return deferred_response

    def http_post(self, url, **kwargs):
        # NOTE
        # The previous implementation returned parsed JSON. This one just
        # returns, and the caller must parse the response as needed.
        self.log.debug("Executing HTTP Post Request\n"
                       " to URL {url}\n"
                       " with kwargs {kwargs}",
                       url=url, kwargs=kwargs)
        deferred_response = self.http_semaphore.run(
            self.http_client.post, url, **kwargs
        )
        deferred_response.addCallbacks(
            self._http_check_response,
            partial(self._http_error_handler, url=url)
        )
        return deferred_response

    def http_download(self, url, dst, semaphore=None, **kwargs):
        if not semaphore:
            semaphore = self.http_semaphore
        deferred_response = semaphore.run(
            self._http_download, url, dst, **kwargs
        )
        return deferred_response

    def _http_download(self, url, dst, **kwargs):
        dst = os.path.abspath(dst)
        self.log.debug("Starting download {url} to {destination}",
                       url=url, destination=dst)
        if os.path.isdir(dst):
            fname = os.path.basename(urlparse(url).path)
            dst = os.path.join(dst, fname)

        if not os.path.exists(os.path.split(dst)[0]):
            os.makedirs(os.path.split(dst)[0])

        self.busy_set()

        _clear_partial_file = None
        if os.path.exists(dst + '.partial'):
            csize = os.path.getsize(dst + '.partial')
            deferred_response = self.http_client.get(
                url, headers={'Range': 'bytes={0}-'.format(csize)}, **kwargs
            )
            _clear_partial_file = dst + '.partial'
        else:
            deferred_response = self.http_client.get(url, **kwargs)

        deferred_response.addCallback(self._http_check_response)
        deferred_response.addErrback(self._deferred_error_passthrough)

        deferred_response.addCallback(
            self._http_download_response, destination_path=dst
        )
        deferred_response.addErrback(
            partial(self._http_error_handler, url=url)
        )

        if _clear_partial_file:
            # If a range request resulted in an error, get rid of the partial
            # file so it'll work the next time
            def _eb_clear_partial_file(failure):
                os.remove(_clear_partial_file)
                return failure
            deferred_response.addErrback(_eb_clear_partial_file)

        def _busy_clear(maybe_failure):
            self.busy_clear()
            return maybe_failure
        deferred_response.addBoth(_busy_clear)

        return deferred_response

    def _http_download_response(self, response, destination_path):
        if response.code == 206:
            # TODO Check that the range is actually correct?
            # self.log.debug("Got partial content response for {dst}",
            #                dst=destination_path)
            append = True
        else:
            # self.log.debug("Got full content response for {dst}",
            #                dst=destination_path)
            append = False
        temp_path = destination_path + '.partial'
        if not append:
            destination = open(temp_path, 'wb')
        else:
            destination = open(temp_path, 'ab')
        collectmethod = partial(watchful_collect, chunktimeout=10, reactor=self.reactor)
        d = collectmethod(response, destination.write)

        def _close_download_file(maybe_failure):
            destination.close()
            return maybe_failure
        d.addBoth(_close_download_file)

        def _finalize_successful_download(_):
            os.rename(temp_path, destination_path)
        d.addCallback(_finalize_successful_download)

        return d

    def _http_error_handler(self, failure, url=None):
        self.log.failure("HTTP Connection Failure to {url} : ", failure=failure, url=url)
        failure.trap(HTTPError, DNSLookupError, ResponseNeverReceived, SchemeNotSupported)
        if isinstance(failure.value, HTTPError):
            self.log.warn(
                "Encountered error {e} while trying to {method} {url}",
                e=failure.value.response.code, url=url,
                method=failure.value.response.request.method
            )
        if isinstance(failure.value, DNSLookupError):
            self.log.warn(
                "Got a DNS lookup error for {url}. Check your URL and "
                "internet connection.", url=url
            )
        if isinstance(failure.value, ResponseNeverReceived):
            self.log.warn(
                "Response never received for {url}. Underlying error is {e}",
                url=url, e=failure.value.reasons
            )
        elif isinstance(failure.value, ResponseFailed):
            self.log.warn(
                "Response Failed for {url}. Underlying error is {e}",
                url=url, e=failure.value.reasons
            )
        if isinstance(failure.value, SchemeNotSupported):
            self.log.warn(
                "Got an unsupported scheme for {url}. Check your URL and "
                "try again.", url=url
            )
        return failure

    def _http_check_response(self, response):
        if 400 < response.code < 600:
            self.log.info("Got a HTTP Error\n"
                          "with Response Code : {code}\n"
                          "and Response Headers : {headers}\n",
                          code=response.code, headers=response.headers)

            d = response.content()

            def _log_error_content(r):
                self.log.info("  Got Response Content : {msg}", msg=r)
            d.addCallback(_log_error_content)

            raise HTTPError(response=response)
        return response

    @property
    def http_semaphore(self):
        if self._http_semaphore is None:
            n = self.config.http_max_concurrent_requests
            self._http_semaphore = DeferredSemaphore(n)
            _ = self.http_client
        return self._http_semaphore

    @property
    def http_semaphore_background(self):
        if self._http_semaphore_background is None:
            n = self.config.http_max_background_downloads
            self._http_semaphore_background = DeferredSemaphore(n)
        return self._http_semaphore_background

    @property
    def http_semaphore_download(self):
        if self._http_semaphore_download is None:
            n = self.config.http_max_concurrent_downloads
            self._http_semaphore_download = DeferredSemaphore(n)
        return self._http_semaphore_download

    @property
    def http_client(self):
        if not self._http_client:
            self.log.info("Creating treq HTTPClient")
            # Silence the twisted.web.client._HTTP11ClientFactory
            from twisted.web.client import _HTTP11ClientFactory
            _HTTP11ClientFactory.noisy = False
            if self.config.http_proxy_enabled:
                proxy_endpoint = TCP4ClientEndpoint(self.reactor,
                                                    self.config.http_proxy_host,
                                                    self.config.http_proxy_port)
                agent = ProxyAgent(proxy_endpoint)
                if self.config.http_proxy_user:
                    auth = base64.b64encode(self.config.http_proxy_auth)
                    self._http_headers['Proxy-Authorization'] = ["Basic {0}".format(auth.strip())]
            else:
                agent = Agent(reactor=self.reactor)
            self._http_client = DefaultHeadersHttpClient(agent=agent, headers=self._http_headers)
        return self._http_client

    def stop(self):
        self.log.debug("Closing HTTP client session")
        super(HttpClientMixin, self).stop()





from twisted.internet import reactor
from ebs.linuxnode.core.basenode import BaseIoTNode
from ebs.linuxnode.core import config


class ExampleNode(BaseIoTNode):
    def start(self):
        self.install()
        super(ExampleNode, self).start()
        self.config.print()
        reactor.callLater(10, self.stop)
        reactor.run()

    def stop(self):
        super(ExampleNode, self).stop()
        reactor.stop()


def main():
    nodeconfig = config.IoTNodeConfig('iotnode-bare')
    config.current_config = nodeconfig

    node = ExampleNode(reactor=reactor)
    node.start()


if __name__ == '__main__':
    main()




# foundation    ()
# backdrop      1
# background    2
# backdrop      3
# video         4
# app           5


import os
import pkg_resources
from six.moves.configparser import ConfigParser
from collections import namedtuple
from appdirs import user_config_dir
from configparser import NoSectionError


ItemSpec = namedtuple('ItemSpec', ["item_type", "fallback", "read_only"], defaults=[str, '_required', True])
ElementSpec = namedtuple('ElementSpec', ['section', 'item', 'item_spec'], defaults=[ItemSpec()])


class IoTNodeConfig(object):
    def __init__(self, appname=None, packagename=None):
        self._elements = {}
        self._packagename = packagename
        self._appname = appname or 'iotnode'
        _root = os.path.abspath(os.path.dirname(__file__))
        self._roots = [_root]
        self._config = ConfigParser()
        print("Reading Config File {}".format(self._config_file))
        self._config.read(self._config_file)
        print("EBS IOT Linux Node Core, version {0}".format(self.linuxnode_core_version))
        self._config_init()

    @property
    def appname(self):
        return self._appname

    @property
    def _config_file(self):
        return os.path.join(user_config_dir(self.appname), 'config.ini')

    @property
    def linuxnode_core_version(self):
        return pkg_resources.get_distribution('ebs-linuxnode-core').version

    @property
    def app_version(self):
        if not self._packagename:
            return
        return pkg_resources.get_distribution(self._packagename).version

    def _write_config(self):
        with open(self._config_file, 'w') as configfile:
            self._config.write(configfile)

    def _check_section(self, section):
        if not self._config.has_section(section):
            self._config.add_section(section)
            self._write_config()

    def _parse_color(self, value, on_error='auto'):
        color = value.split(':')
        if len(color) not in (3, 4):
            return on_error
        try:
            color = (float(x) for x in color)
        except ValueError:
            return on_error
        return tuple(color)

    # Paths
    @property
    def roots(self):
        return list(reversed(self._roots))

    def get_path(self, filepath):
        if not filepath:
            return filepath
        for root in self.roots:
            if os.path.exists(os.path.join(root, filepath)):
                return os.path.join(root, filepath)
        return filepath

    def register_application_root(self, root):
        self._roots.append(root)

    # Modular Config Infrastructure
    def register_element(self, name, element_spec):
        self._elements[name] = element_spec

    def __getattr__(self, element):
        if element not in self._elements.keys():
            raise AttributeError(element)
        section, item, item_spec = self._elements[element]
        item_type, fallback, read_only = item_spec
        kwargs = {}
        if not fallback == "_required":
            kwargs['fallback'] = fallback
        if section == '_derived':
            return item(self)
        if item_type == str:
            return self._config.get(section, item, **kwargs)
        elif item_type == bool:
            return self._config.getboolean(section, item, **kwargs)
        elif item_type == int:
            return self._config.getint(section, item, **kwargs)
        elif item_type == float:
            return self._config.getfloat(section, item, **kwargs)
        elif item_type == 'kivy_color':
            return self._parse_color(self._config.get(section, item, **kwargs))
        elif item_type == 'path':
            return self.get_path(self._config.get(section, item, **kwargs))

    def __setattr__(self, element, value):
        if element == '_elements' or element not in self._elements.keys():
            return super(IoTNodeConfig, self).__setattr__(element, value)
        section, item, item_spec = self._elements[element]
        item_type, fallback, read_only = item_spec

        if read_only:
            raise AttributeError("{} element '{}' is read_only. Cannot write."
                                 "".format(self.__class__.__name__, element))
        if item_type == bool:
            value = "yes" if value else "no"

        if not value:
            value = ''

        self._check_section(section)
        self._config.set(section, item, value)
        self._write_config()

    def remove(self, element):
        section, item, item_spec = self._elements[element]
        if item_spec.read_only:
            return False
        try:
            return self._config.remove_option(section, item)
        except NoSectionError:
            pass

    def _config_init(self):
        _elements = {
            'platform': ElementSpec('platform', 'platform', ItemSpec(fallback='native')),
        }

        for element, element_spec in _elements.items():
            self.register_element(element, element_spec)

    def print(self):
        print("Node Configuration ({})".format(self.__class__.__name__))
        for element in self._elements.keys():
            print("    {:>30}: {}".format(element, getattr(self, element)))


class ConfigMixin(object):
    def __init__(self, *args, **kwargs):
        global current_config
        self._config: IoTNodeConfig = current_config
        super(ConfigMixin, self).__init__(*args, **kwargs)

    @property
    def config(self):
        return self._config

    def config_register_element(self, name, element_spec):
        self.config.register_element(name, element_spec)
